/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------
*/

#ifndef _UCONFIG_H_
#define _UCONFIG_H_

#include "ulib.h"
#include "uarray.h"

#define MAXLINE	2048

#define KB 1024
#define MB (KB*KB)
#define GB (MB*KB)

int isCommentLine(char *l)
{
return l[0] == '%' || l[0] == '#' || l[0] == 0;
}

char *readALine(char *line, int len, FILE *fp)
{
if (feof(fp))
    return NULL;

int rl=1;

do
    {
    if (!fgets(line, len, fp))
	rl = 0;
    }while (!feof(fp) && isCommentLine(line)); 
if (rl && !isCommentLine(line))
    {
    int l=strlen(line);
    if (line[l-1]=='\n')
	line[l-1] = 0;
    if (l==len-1)
	fprintf(stderr, "line is too long, re-compile with larger MAXLINE(%d)",len);
    else
	return utrim(line);
    }
return NULL;
}


int getSuffixMultiplier(char *st)
{
 int l;

 if (!st)
     return 0;
 l=strlen(st);
            
 if (l>2 && toupper(st[l-1])=='B')
     l = l-2;
 else if (l>1)
     l = l-1;
 else
     l = -1;
            
 if (l!=-1)
     {
     char c=toupper(st[l]);
                
     if (c=='G')
         return GB;
     else if (c=='M')
         return MB;
     else if (c=='K')
         return KB;                
     }
 return 1;
}


class UConfigEntry
    {
    private:
        char *key, *value;
        
    public:
	UConfigEntry(FILE *fp, char *line, int len) : key(NULL), value(NULL)
	    {
	    if (readALine(line, len, fp))
		{
		char *p=strchr(line, '=');
		if (p)
		    {
		    *p = 0;
		    key=strdup(utrim(line));
		    value=strdup(utrim(p+1));
		    }
		}
	    }
	~UConfigEntry()
	    {
	    if (key)
		free(key);
	    if (value)
		free(value);
	    }

	bool isValid()		{ return key!=NULL; }

	char *getKey()		{ return key; }
	char *getValue()	{ return value; }
	operator char*()	{ return getValue(); }
        
	operator int()          { return atoi(*this)*getSuffixMultiplier(value); }
	operator double()	{ return atof(*this)*getSuffixMultiplier(value); }

	friend std::ostream& operator<<(std::ostream& os, UConfigEntry &e)
	    {
	    os << e.getKey() << "=" << e.getValue() << std::endl;
	    return os;
	    }
    };


class UConfigSection
    {
    private:
	UArray<UConfigEntry *> entries;
	char  *name;

    public:
	UConfigSection(FILE *fp, char *line, int len) : name(NULL)
	    { 
	    UConfigEntry *e;

	    if (line[0]!='[')
		{
		do
		    {
		    if (!readALine(line, len, fp))
			return;
		    } while (!feof(fp) && isCommentLine(line));
		}
	    if (!feof(fp))
		{
		int l=strlen(line);
		if (line[0] != '[' || line[l-1] != ']')
			errexit("UConfigSection::read: was expecting '[...]' but found '%s'\n", line);
		line[l-1] = 0;
		name = strdup(line+1);
		do
		    {
		    e = new UConfigEntry(fp, line, len);
		    if (e->isValid())
			entries.append(e);
		    else
			delete e;
		    } while (!feof(fp) && line[0] != '[');
		}
	    }
	~UConfigSection()
	    {
	    entries.deleteAll();
	    }
	
	bool isValid()			    { return name!=NULL; }
	char *getName()			    { return name; }
	UConfigEntry* operator[](int i)	    { return entries[i]; }
	int getNEntries()		    { return entries.getSize(); }

	friend std::ostream& operator<<(std::ostream& os, UConfigSection &sec)
	    {
	    os << '[' << sec.getName() << ']' << std::endl;
	    for (int i=0; i<sec.getNEntries(); ++i)
		os << *sec[i];
	    return os;
	    }

	UConfigEntry *entry(char *key)
	    {
	    UConfigEntry *e=getEntry(key);
	    if (!e)
		errexit("UConfigEntry::entry failed for key='%s'\n", key);
	    return e;
	    }
	UConfigEntry *getEntry(char *key)
	    {
	    int i;
	    for (i=0; i<getNEntries() && strcmp(key, entries[i]->getKey()); ++i);
	    if (i<getNEntries())
		return entries[i];
	    return NULL;
	    }
    };

class UConfig
    {
    private:
	UArray<UConfigSection *>		sections;
	int cursection;

    public:
	UConfig(char *configfile) : cursection(0)
	    {
	    char line[MAXLINE];
	    UConfigSection *sec;

	    line[0] = 0;
	    FILE *fp=fopen(configfile, "r");
	    if (fp==NULL)
		return;
	    do
		{
		sec = new UConfigSection(fp, line, sizeof(line));
		if (sec->isValid())
		    sections.append(sec);
		else
		    delete sec;
		} while (!feof(fp));
	    fclose(fp);
	    }
        ~UConfig()
            {
            sections.deleteAll();
            }
        

	UConfigSection  *operator[](int i)	    { return sections[i]; }
	int getNSections()			    { return sections.getSize(); }

	friend std::ostream& operator<<(std::ostream& os, UConfig &conf)
	    {
	    for (int i=0; i<conf.getNSections(); ++i)
		{
		os << *conf[i];
		if (i<conf.getNSections()-1)
		    os << std::endl;
		}
	    return os;
	    }

	UConfigSection *current() { return (cursection<getNSections()) ? sections[cursection] : NULL; }
	UConfigSection *first()   { cursection = 0; return current(); }
	UConfigSection *next()    { ++cursection; return current(); }
	
	UConfigSection *firstSection(char *secname)
	    {
	    for ( ; cursection<getNSections() && strcmp(secname, sections[cursection]->getName()); ++cursection);
	    if (cursection<getNSections())
                return sections[cursection];
	    return NULL;
	    }

	UConfigSection *nextSection(char *secname)
	    {
	    next();
	    return firstSection(secname);
	    }

	UConfigSection *firstSec(char *secname)
	    {
	    UConfigSection *s=firstSection(secname);
	    if (!s)
                errexit("UConfig::firstSec: there is no section named '%s'", secname);
	    return s;
	    }
    };


#if 0
#undef KB 
#undef MB 
#undef GB
#endif

#endif

