/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------

  /-------------------------------------------------------------\
  |                     File : heap.h                           |
  +-------------------------------------------------------------+
  |           Written by : Umit V. CATALYUREK                   |
  +-------------------------------------------------------------+
  |           Date       : 03.5.97                              |
  +-------------------------------------------------------------+
  |  Description : Heap                                         |
  +-------------------------------------------------------------+
  |  Modification by :                                          |
  |  Date            :                                          |
  |  Description     :                                          |
  \-------------------------------------------------------------/
*/
#ifndef _HEAP_H_
#define _HEAP_H_

#include "ulib.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef int		HeapValType;
typedef unsigned int    UnSignedHeapValType;

/* --------------------- Heap : Data structures ------------------------ */
typedef struct heap
{
int		    *data;
int		    *ptr;	    /* index of id in data */
HeapValType	    *vals;          /* associated values of ids (id is index of
				       this array */
UnSignedHeapValType *usvals;
int		    size;
int		    maxid,
		    cmaxid;
} heap;

typedef heap        *pheap;


/*************************************************************************
* ------------------------------- Macros ---------------------------------
**************************************************************************/
/* generic macros applicable to all heaps */
#define u_IsHeapEmpty(h)	  ((h)->size==0)
#define u_HeapGetSize(h)          ((h)->size)

/* macros applicable to signed heaps */
#define u_HeapGetVal(h, id)       ((h)->vals[id])
#define u_HeapGetFirstVal(h)      ((h)->vals[(h)->data[1]])

/* macros applicable to unsigned heaps */
#define u_UnSigned_HeapGetVal(h, id)       ((h)->usvals[id])
#define u_UnSigned_HeapGetFirstVal(h)      ((h)->usvals[(h)->data[1]])

/* macros applicable to both signed and unsigned heaps */
#define u_IsInHeap(h, id)         ((h)->ptr[id]!=-1)

/* macros applicable to tripleheaps */
#define u_TripleHeapGetFirstVal(h)      ((h)->data[1].val)


/*************************************************************************
* --------------------------- Heap Functions -----------------------------
**************************************************************************/

/***** General heap functions applicable to all heaps except tripleheap *****/
int  u_HeapAlloc(pheap h, int Imaxid);
void u_HeapInit(pheap h, int Imaxid);
void u_HeapFree(pheap h);
void u_HeapDeleteAll(pheap h);

void u_HeapPrint(pheap h, char *name);
int u_HeapVerify(pheap h, char *st);

/* Note that this add does not sort heap You must explicitly call Heapify */
void u_HeapAdd(pheap h, int id, HeapValType v);	

/*************************************************************************
* --------------------- Generalized Heap Functions -----------------------
 Desr: Use with normal heap structure, but it requireds to give
 comparison function (use to store multiple value, e.g. 2 words or string
**************************************************************************/

/** if gen_heap_compare gives
  > 0  for val(id1) > val(id2)
 == 0  for val(id1) == val(id2)
  < 0  for val(id1) < val(id2)
 >>>>> then heap is min heap 
 >>>>> otherwise max heap
 ***/
typedef int (*gen_heap_compare)(int id1, int id2);

void u_Gen_HeapAdd(pheap h, int id);


void u_Gen_HeapInsert(pheap h, int id, gen_heap_compare func);
void u_Gen_Heapify(pheap h, gen_heap_compare func);
int  u_Gen_HeapExtractFirst(pheap h, gen_heap_compare func);
void u_Gen_HeapDelete(pheap h, int id, gen_heap_compare func);

void u_Gen_HeapModifiedAt(pheap h, int p, gen_heap_compare func);

void u_Gen_HeapUpdateVal(pheap h, int id, gen_heap_compare func);


/*************************************************************************
* --------------------- signed MinHeap Functions -------------------------
**************************************************************************/
void u_MinHeapInsert(pheap h, int id, HeapValType v);
void u_MinHeapify(pheap h);
int  u_MinHeapExtractFirst(pheap h);
void u_MinHeapDelete(pheap h, int id);

void u_MinHeapModifiedAt(pheap h, int p);

void u_MinHeapUpdateVal(pheap h, int id, HeapValType v);


/*************************************************************************
* --------------------- signed MaxHeap Functions -------------------------
**************************************************************************/
void u_MaxHeapInsert(pheap h, int id, HeapValType v);
void u_MaxHeapify(pheap h);
int  u_MaxHeapExtractFirst(pheap h);
void u_MaxHeapDelete(pheap h, int id);

#define u_MaxHeapSafeDelete(h, id) { if (u_IsInHeap((h), (id))) u_MaxHeapDelete((h), (id)); }


void u_MaxHeapModifiedAt(pheap h, int p);

void u_MaxHeapUpdateVal(pheap h, int id, HeapValType v);

/*************************************************************************
* --------------------- unsigned MaxHeap Functions -----------------------
**************************************************************************/
/* Note that this add does not sort heap You must explicitly call Heapify */
void u_UnSigned_HeapAdd(pheap h, int id, UnSignedHeapValType v);

void u_UnSigned_MaxHeapInsert(pheap h, int id, UnSignedHeapValType v);
void u_UnSigned_MaxHeapify(pheap h);
int  u_UnSigned_MaxHeapExtractFirst(pheap h);
void u_UnSigned_MaxHeapDelete(pheap h, int id);

void u_UnSigned_MaxHeapModifiedAt(pheap h, int p);

void u_UnSigned_MaxHeapUpdateVal(pheap h, int id, UnSignedHeapValType v);

#ifdef __cplusplus
}
#endif

#endif
