/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------

  /-------------------------------------------------------------\
  |                     File : ulib.h                           |
  +-------------------------------------------------------------+
  |           Written by : Umit V. CATALYUREK                   |
  +-------------------------------------------------------------+
  |           Date       : 27.9.95                              |
  +-------------------------------------------------------------+
  |  Description : Some useful library functions, such as;      |
  |                                                             |
  |     1. File open/close                                      |
  |     2. Timer functions                                      |
  |     3. Safe memory allocation functions                     |
  |     4. Sorting functions                                    |
  |     5. Safe string functions                                |
  |     6. Nice error exit function                             |
  +-------------------------------------------------------------+
  |  Modification by :                                          |
  |  Date            :                                          |
  |  Description     :                                          |
  \-------------------------------------------------------------/
*/
#ifndef _ulib_H_
#define _ulib_H_

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <fcntl.h>

#ifdef __cplusplus
extern "C" {
#endif

#include "uport.h"

    
/*
#define _DEBUGPOOL
*/

/*************************************************************************
* --------------------- Some Useful Types --------------------------------
**************************************************************************/
typedef int		   *pint;
typedef char		   *pchar;
typedef float		   *pfloat;
typedef double		   *pdouble;
typedef void               *pvoid;



/*************************************************************************
* --------------------- Some Useful Macros -------------------------------
**************************************************************************/

/* maximum representable numbers */ 
#define U_MAXINT        2000000000
#define U_MAXDOUBLE	1e+300

/* epsilons */
#define u_fEPSILON	1e-6
#define u_dEPSILON	1e-12


/* booleans */
#define NONE		0
#define NO		0
#define YES		1

#define FALSE		0
#define TRUE		1


#define log2(x)		((int) (log((double)(x)) / log(2.0)))
#define umax(a, b)	(((a) >= (b)) ? (a) : (b))
#define umin(a, b)	(((a) >= (b)) ? (b) : (a))
#define urange(l, u, x) (umin((u), umax((l), (x))))
#define uabs(a)		(((a) < 0) ? (-(a)) : (a))
#define usqr(x)		((x)*(x))
#define unearest(x)	(((x)-floor(x) < (ceil(x) - (x))) ? floor(x) : ceil(x))

#define u_fNEARZERO(x)   (uabs(x) < u_fEPSILON)
#define u_dNEARZERO(x)   (uabs(x) < u_dEPSILON)



extern int _u_seed;

#define usRandom(s)     _u_seed = s
#define usRandomize     usRandom((int)time(NULL))


    
#ifdef NO_INLINE
    int uGetSeed(void);
    int uRandom(int l);
#else
UINLINEKEYWORD int uGetSeed(void)
{
return _u_seed;
}

UINLINEKEYWORD int uRandom(int l)
{
int   a = 16807;
int   m = 2147483647;
int   q = 127773;
int   r = 2836;
int   lo, hi, test;
double d;
 
lo = _u_seed % q;
hi = _u_seed / q;
test = (a*lo)-(r*hi);
if (test>0)
    _u_seed = test;
else
    _u_seed = test + m;
d = (double) ((double) _u_seed / (double) m);
return (int) (d*(double)l);
}
#endif

/*************************************************************************
* --------------------------- Useful -------------------------------------
**************************************************************************/
#define bzero(ptr,size)  memset((void*)(ptr), '\0', (size))

void uWaitKbdNewLine(void);
void uprintf(pchar fmt,...);
void udbgprintf_SetDbgFile(char *name); /* pass null to write default debug file */
void udbgprintf(pchar f_str,...); /* prints only debug file */


int isThisMachineBigEndian(void);

/* non conditional swaps */
#define SWAP_BYTES_NC(v,l) \
{ \
char *_bs,*_be; \
char _btemp; \
_bs = ((char*)v); _be = ((char*)v)+l-1; \
while (_bs<_be) { \
_btemp = *_bs; *_bs = *_be; *_be = _btemp; \
_bs++; _be--; \
} \
}

void SwapByteArray_NC(void *p, int elemsz, int numelem);


#ifdef _U_DO_ENDIAN_CONVERSION

#define SWAP_BYTES(v,l)           SWAP_BYTES_NC(v,l)
#define SwapByteArray(p, sz, n)   SwapByteArray_NC(p, sz, n)

#else 

#define SWAP_BYTES(v,l)
#define SwapByteArray(p, sz, n)

#endif 


/*************************************************************************
* --------------------------- Time ---------------------------------------
**************************************************************************/

/* --------------------- Time : Constants ------------------------------ */

#define TMR_CLEAR   1
#define TMR_START   2
#define TMR_STOP    8

/* --------------------- Time : Data structures ------------------------ */

struct timer 
    {
    int    status;    /* current status of timer look TMR_xxxx constants */
    double time;      /* time in seconds */
    long   count;     /* counts how many times the counter was started */
    };

typedef struct timer    timer;
typedef timer           *ptimer;

/* --------------------- Time : Functions ------------------------------ */

void u_cleartimer(ptimer);
void u_starttimer(ptimer);
void u_addtimer(ptimer t, ptimer a);
void u_stoptimer(ptimer);
void u_printtimer(ptimer, pchar msg);
char *u_gettimestr(ptimer tmr);
double u_gettimer(ptimer);
double u_safegettimer(ptimer);    


struct wtimer 
    {
    int    status;    /* current status of timer look TMR_xxxx constants */
    double time;      /* time in seconds */
    long   count;     /* counts how many times the counter was started */
    };

typedef struct wtimer       wtimer;
typedef wtimer              *pwtimer;

/* --------------------- Time : Functions ------------------------------ */

void u_clearwtimer(pwtimer);
void u_startwtimer(pwtimer);
void u_addwtimer(pwtimer t, pwtimer a);
void u_stopwtimer(pwtimer);
void u_printwtimer(pwtimer, pchar msg);
char *u_getwtimestr(pwtimer tmr);
double u_getwtimer(pwtimer);


/*************************************************************************
* --------------------------- File ---------------------------------------
**************************************************************************/

/* --------------------- File : Functions ------------------------------ */

FILE *ufopen(pchar fname, pchar mode, pchar msg);
void ufclose(FILE *);

int  uopen(pchar fname, int flags, pchar msg);
void uclose(int fid);

/*************************************************************************
* -------------------------- Memory --------------------------------------
**************************************************************************/

/* --------------------- Memory : Functions ---------------------------- */
void ReportAllocMem(void);
void DetailedReportAllocMem(void);

#ifdef NO_INLINE
void iset(pint a, int to, int sz);
void izero(pint a, int sz);
void iadd(pint sum, pint add, int sz);
void isub(pint sum, pint sub, int sz);
int minindexof(pint lst, int size);
int maxindexof(pint lst, int size);
int minof(pint lst, int size);
int maxof(pint lst, int size);
int sumof(pint lst, int size);
void iswap(pint p1, pint p2);
void icopy(pint dest, pint src, int size);

/* float versions */
void fset(pfloat a, float to, int sz);
void fzero(pfloat a, int sz);
void fadd(pfloat sum, pfloat add, int sz);
void fsub(pfloat sum, pfloat sub, int sz);
int fminindexof(pfloat lst, int size);
int fmaxindexof(pfloat lst, int size);
float fminof(pfloat lst, int size);
float fmaxof(pfloat lst, int size);
float fsumof(pfloat lst, int size);
void fcopy(pfloat dest, pfloat src, int size);

#else
UINLINEKEYWORD void iset(pint a, int to, int sz)
    {
    int i;

    for (i=0; i<sz; ++i)
	a[i] = to;
    }

UINLINEKEYWORD void izero(pint a, int sz)
    {
    memset(a, 0, sizeof(int)*sz);
    }


UINLINEKEYWORD void iadd(pint sum, pint add, int sz)
    {
    int i;

    for (i=0; i<sz; ++i)
	sum[i] += add[i];
    }

UINLINEKEYWORD void isub(pint sum, pint sub, int sz)
    {
    int i;

    for (i=0; i<sz; ++i)
	sum[i] -= sub[i];
    }


UINLINEKEYWORD int minindexof(pint lst, int size)
    {
    int i, min;

    min = 0;
    for (i=1; i<size; i++)
	if (lst[i]<lst[min])
	    min = i;
    return min;
    }

UINLINEKEYWORD int maxindexof(pint lst, int size)
    {
    int i, max;

    max = 0;
    for (i=1; i<size; i++)
	if (lst[i]>lst[max])
	    max = i;
    return max;
    }


UINLINEKEYWORD int minof(pint lst, int size)
    {
    int i, min;

    min = lst[0];
    for (i=1; i<size; i++)
	if (lst[i]<min)
	    min = lst[i];
    return min;
    }

UINLINEKEYWORD int maxof(pint lst, int size)
    {
    int i, max;

    max = lst[0];
    for (i=1; i<size; i++)
	if (lst[i]>max)
	    max = lst[i];
    return max;
    }

UINLINEKEYWORD int sumof(pint lst, int size)
    {
    int i, sum=0;

    for (i=0; i<size; i++)
	sum += lst[i];
    return sum;
    }

UINLINEKEYWORD void iswap(pint p1, pint p2)
    {
    int t=*p1;
    *p1 = *p2;
    *p2 = t;
    }

UINLINEKEYWORD void icopy(pint dest, pint src, int size)
    {
    memcpy((pvoid )dest, (pvoid )src, sizeof(int)*size);
    }


/* float versions */

UINLINEKEYWORD void fset(pfloat a, float to, int sz)
    {
    int i;

    for (i=0; i<sz; ++i)
	a[i] = to;
    }

UINLINEKEYWORD void fzero(pfloat a, int sz)
    {
    fset(a, 0.0, sz);
    }

UINLINEKEYWORD void fadd(pfloat sum, pfloat add, int sz)
    {
    int i;

    for (i=0; i<sz; ++i)
	sum[i] += add[i];
    }

UINLINEKEYWORD void fsub(pfloat sum, pfloat sub, int sz)
    {
    int i;

    for (i=0; i<sz; ++i)
	sum[i] -= sub[i];
    }

UINLINEKEYWORD int fminindexof(pfloat lst, int size)
    {
    int i, min;

    min = 0;
    for (i=1; i<size; i++)
	if (lst[i]<lst[min])
	    min = i;
    return min;
    }

UINLINEKEYWORD int fmaxindexof(pfloat lst, int size)
    {
    int i, max;

    max = 0;
    for (i=1; i<size; i++)
	if (lst[i]>lst[max])
	    max = i;
    return max;
    }


UINLINEKEYWORD float fminof(pfloat lst, int size)
    {
    int   i;
    float min;

    min = lst[0];
    for (i=1; i<size; i++)
	if (lst[i]<min)
	    min = lst[i];
    return min;
    }

UINLINEKEYWORD float fmaxof(pfloat lst, int size)
    {
    int   i;
    float max;

    max = lst[0];
    for (i=1; i<size; i++)
	if (lst[i]>max)
	    max = lst[i];
    return max;
    }

UINLINEKEYWORD float fsumof(pfloat lst, int size)
    {
    int   i;
    float sum=0;

    for (i=0; i<size; i++)
	sum += lst[i];
    return sum;
    }


UINLINEKEYWORD void fcopy(pfloat dest, pfloat src, int size)
    {
    memcpy((pvoid )dest, (pvoid )src, sizeof(float)*size);
    }

#endif 


pint imalloc(long size, pchar msg);
pfloat fmalloc(long size, pchar msg);
pdouble dmalloc(long size, pchar msg);
pvoid umalloc(long size, pchar msg);
pvoid urealloc(pvoid ptr, long size, pchar msg);
void ufree(pvoid ,...);  /* IMPORTANT : Last parameter must be NULL */
void sicopy(pint dest, pint src, int size); /* safe icopy */
pint ismalloc(long size, pchar msg);
pfloat fsmalloc(long size, pchar msg);
pdouble dsmalloc(long size, pchar msg);
pvoid usmalloc(long size, pchar msg);
pvoid usrealloc(pvoid ptr, long size, pchar msg);

#define usfree(x)     do { _usfree(x); (x) = NULL; } while (0)
void _usfree(pvoid );    /* Safe free checks if parameter is not null*/
    
void unfree(int n, ...); /* to free n ptrs: BUT pass address of pointers
                            because it sets them to NULL after free
                            to be safe */
    
/* allocation pool indicies*/ 
#define POOL_CELLNET	0
#define POOL_PINS	1
#define POOL_GENERAL    2	

#define MAXALLOCPOOL    3

typedef struct _AllocPoolInfo
    {
    pchar          base, ptr, last; 
    int            isallocated;
    long	   maxused;
    } AllocPoolInfo;

typedef AllocPoolInfo AllocationPools[MAXALLOCPOOL];

void InitAllocationPools(void);
int  IsPoolAllocated(int pool);
int AllocatePool(int pool, long size, char *msg);
int sAllocatePool(int pool, long size, char *msg);
void ReInitPools(void);
void FreePool(int pool);
void FreePools(void);
void ReportMaxPoolUsages(void);
void ShortReportMaxPoolUsages(void);
void SavePools2Check(void);
void CheckPools(void);
pchar poolalloc(int pool, long size, char *msg);
pchar poolgetptr(int pool);
void poolfree(int pool, long size, char *msg);
void poolfree_upto(int pool, pchar ptr, char *msg);
int ispoolavailable(int pool, long size);

#define upalloc(pool, size, msg)    poolalloc(pool, size, msg)
#define upfree(pool, ptr, msg)	    poolfree_upto(pool, (pchar) ptr, msg)
#define updealloc(pool, size, msg)  poolfree(pool, size, msg)


/*************************************************************************
* -------------------------- String --------------------------------------
**************************************************************************/

/* --------------------- String : Functions ---------------------------- */
int uiswhitespace(char c);
pchar ultrim(pchar src);
pchar urtrim(pchar src);
pchar utrim(pchar src);
pchar ustrcpy(pchar dest, pchar src);
pchar ustrncpy(pchar dest, pchar src, long n);
pchar ustrdup(pchar st);
pchar GetTabSpace(int tabcnt);



/* --------------------- Sort : Functions ---------------------------- */
void isorttwo(pint i, pint j);



/*************************************************************************
* -------------------------- Usefuls   ----------------------------------
**************************************************************************/

void uwarning(pchar f_str,...);

#define uassert(c_) \
		do {\
		   if (!(c_)) errexit("ulib: Assertion failed in file '%s' line %d", __FILE__, __LINE__);\
		} while (0)


/* randomly permutes n integers stored in a[0]..a[n-1] */
void PermuteInPlace(pint a, int n);
int ispowerof2(int n);

pint read_array(char *fname, pint resvec, int n, int base);

#ifdef __cplusplus
}

#if 0
void* operator new(size_t sz);
void operator delete(void* m);
#endif

#endif


#endif 
