/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------
    
   ------------------------------------------------------------------
   File   : usocket.c
   Author : Umit V. Catalyurek
   Date   : Nov 2, 2000
   ------------------------------------------------------------------
   Desc   : Socket functions

   ------------------------------------------------------------------
*/
#include <errno.h>
#include "usocket.h"
#ifdef UNIX
#include <netinet/tcp.h>
#endif
#include "ulib.h"


int UInitializeSocketOpt(USocketOptions *sopt)
{
 sopt->nodelay=1;
 sopt->sendbufsz=0;
 sopt->recvbufsz=0;
 sopt->nconnecttry=10;
 return 0;
}


void getdispopt(USOCKET sd, int lev, int opt, char *st)
{
 int		val;
 USOCKLEN   len;

 getsockopt(sd, lev, opt, &val, (USOCKLEN *)&len);
 uprintf("Level=%3d %-15s (%4d) = %7d (len=%4d)\n", lev, st, opt, val, len); 
}


void PrintSocketOptions(USOCKET sd, int lev, char *st) 
{
 uprintf("\n----------------- %s (%d) ----------------\n", st, lev);
 getdispopt(sd, lev, TCP_NODELAY, "TCP_NODELAY");
 getdispopt(sd, lev, SO_SNDBUF, "SO_SNDBUF");
 getdispopt(sd, lev, SO_RCVBUF, "SO_RCVBUF");
 getdispopt(sd, lev, SO_SNDLOWAT, "SO_SNDLOWAT");
 getdispopt(sd, lev, SO_RCVLOWAT, "SO_RCVLOWAT");
 getdispopt(sd, lev, SO_SNDTIMEO, "SO_SNDTIMEO (ro)");
 getdispopt(sd, lev, SO_RCVTIMEO, "SO_RCVTIMEO (ro)");
 getdispopt(sd, lev, SO_TYPE, "SO_TYPE (ro)");
}


void PrintAllSocketOptions(USOCKET sd)
{
 PrintSocketOptions(sd, IPPROTO_TCP, "IPPROTO_TCP");
}


USOCKET _USetupSocket(USOCKET sd, USocketOptions *sopt)
{
 if (!sopt)
     return sd;

/* PrintAllSocketOptions(sd);*/
 
/* 
 struct protoent *proto;
 
 if (!(proto = getprotobyname("tcp"))) 
     errexit("ULib: protocol 'tcp' unknown!\n");
 if (setsockopt(sd, proto->p_proto, TCP_NODELAY, &one, sizeof(one)) < 0)
     */

 if (sopt->nodelay)
     {  
     if (setsockopt(sd, IPPROTO_TCP, TCP_NODELAY, &(sopt->nodelay), sizeof(sopt->nodelay)) < 0) 
         errexit("_USetupSocket: setsockopt: TCP_NODELAY failed! errno=%d\n", errno);
     }

 if (sopt->sendbufsz)
     {
     if(setsockopt(sd, SOL_SOCKET, SO_SNDBUF, &(sopt->sendbufsz), sizeof(sopt->sendbufsz)) < 0)
         errexit("_USetupSocket: setsockopt: SO_SNDBUF failed! errno=\n", errno);
     }

 if (sopt->recvbufsz)
     {      
     if(setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &(sopt->recvbufsz), sizeof(sopt->recvbufsz)) < 0)
         errexit("_USetupSocket: setsockopt: setsockopt: SO_RCVBUF failed! errno=\n", errno);
     }
 
 return sd;
}


USOCKET UConnect_to_Host_wIP(UHOSTIP ip, UPORT port, USocketOptions *sopt)
{
 unsigned long l=htonl(ip);
 struct in_addr adr= *((struct in_addr *) &l);

 char hostname[512];

 strcpy(hostname, inet_ntoa(adr));
 
/* uprintf("\n connect to host '%s' port %d\n", hostname, port); */
 
 return UConnect_to_Host(hostname, port, sopt);
}


USOCKET UConnect_to_Host(char *hostname, UPORT port, USocketOptions *sopt)
{
 USOCKET serversd;
 struct sockaddr_in addr;
 struct hostent *host;
 int na;
 
 
     /* Determine server host address. */
 if ((host = gethostbyname(hostname)) == NULL)
     {
     uprintf("UConnect_to_Host: gethostbyname error\n");
     return INVALID_SOCKET;
     }
 
     /* Initialize server address. */
 bzero((char *)&addr, sizeof(addr));
 addr.sin_family = AF_INET;

/* uprintf("\n DEBUG connect to '%s'\n", inet_ntoa(*((struct in_addr *)host->h_addr))); */
 
 
 addr.sin_addr.s_addr = inet_addr(inet_ntoa(*((struct in_addr *)host->h_addr)));
 addr.sin_port = htons(port);
 
     /* Create socket. */
 if ((serversd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
     {
     uprintf("UConnect_to_Host: socket error\n");
     return INVALID_SOCKET;
     }

 _USetupSocket(serversd, sopt);
     /* And finally connect to the server. */

 for (na=0; na<sopt->nconnecttry; )
     {
     if (connect(serversd, (struct sockaddr *) &addr, sizeof(addr)) < 0)
         {
         uprintf("Host : %s\n", hostname);
         perror("UConnect_to_Host:connect");
         if (++na<sopt->nconnecttry)
             usleep(1000*(na*750)/2+uRandom(na*375));
         }
     else
         break;				/* connect succeeded ! */
     }
 
 if (na >= sopt->nconnecttry) /* too many failures */
     {
     uprintf("Host : %s\n", hostname);
     perror("UConnect_to_Host:connect");
     uprintf("UConnect_to_Host: cannot connect to '%s' Error code is %d. Tried %d times\n", hostname, errno, na);
     close(serversd);
     return INVALID_SOCKET;
     }
 
 return serversd;
}


/* give 0 to start a server at any available port */
USOCKET UServerSocket(UPORT *pport, USocketOptions *sopt)
{
 UPORT   port=*pport;
 USOCKET listensd;
 int     i;
 USOCKLEN len;
 struct  sockaddr_in  addr;
 
  
     /* Create socket. */
 if ((listensd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
     {
     uprintf("UServerSocket: Can't open socket.\n");
     return INVALID_SOCKET;
     }

     /* Bind local address. */
 bzero((char *)&addr, sizeof(addr));
 addr.sin_family = AF_INET;
 addr.sin_addr.s_addr = htonl(INADDR_ANY);
 --port;
 i = 0;
 do
     {
     addr.sin_port = htons(++port);
     if (++i>MAX_PORT_TRY)
          errexit("UServerSocket: I tried %d ports but non of the free", MAX_PORT_TRY);
     } while (bind(listensd, (struct sockaddr *)&addr, sizeof(addr)) < 0);

 len = sizeof(addr);
 if (getsockname(listensd, (struct  sockaddr *) &addr, &len) < 0)
     {
     uprintf("UServerSocket: getsockname failed.\n");
     closesocket(listensd);
     return INVALID_SOCKET;
     }
 
 *pport = ntohs(addr.sin_port);
 _USetupSocket(listensd, sopt);
     /* Now start listening to connection requests. */
 if (listen(listensd, 51)<0)
     {
     uprintf("UServerSocket: listen failed.\n");
     closesocket(listensd);
     return INVALID_SOCKET;    
     }
 return listensd;
}

int UWaitAndAcceptAr_Socket(USOCKET *sl, int nsocket, int *acsl, USocketOptions *sopt)
{
 fd_set	                fdset;
 int                    sel, i, width=0, accepted=0;
 struct sockaddr_in     addr;
 USOCKLEN              len=sizeof(addr);
 
 FD_ZERO(&fdset);
 for (i=0; i<nsocket; ++i)
     {
     if (sl[i] != INVALID_SOCKET)
         {
         FD_SET(sl[i], &fdset);     
         if ((int)sl[i] > width) 
             width = sl[i];
         }
     }
 sel = select(width+1, &fdset, NULL, NULL, NULL);
 if (!sel)
     errexit("UWaitAndAccept_Socket: select failed with sel=0");

 for (i=0; i<nsocket; ++i)
     if ((sl[i] != INVALID_SOCKET) && FD_ISSET(sl[i], &fdset))
         {
         acsl[accepted] = accept(sl[i], (struct sockaddr *)&addr, &len);
         if (acsl[accepted] < 0)
             uprintf("UWaitAndAccept_Socket: Error in accepting connection\n");
         else
             _USetupSocket(acsl[accepted], sopt);
         ++accepted;
         }

/* uprintf("UWaitAndAccept_Socket: nsocket=%d *pidx=%d newsd=%d sel=%d sweep=%d\n", nsocket, *pidx, newsd, sel, sweep); */
 
 return accepted;
}



UHOSTIP UGetLocalHostIP()
{
 struct hostent *host;
 char lochost[256];

 gethostname(lochost, sizeof(lochost));
 if (!(host=gethostbyname(lochost)))
     errexit("UGetLocalHostIP: gethostbyname(%s) failed!", lochost);

/* uprintf("\n lochost='%s' or '%s' IP = %x\n", lochost, inet_ntoa(*((struct in_addr *)host->h_addr)), ntohl(inet_addr(inet_ntoa(*((struct in_addr *)host->h_addr))))); */
 
 return ntohl(inet_addr(inet_ntoa(*((struct in_addr *)host->h_addr))));
}


UHOSTIP UGetHostIP(char *hostname)
{
 struct hostent *host;

 if (!(host=gethostbyname(hostname)))
     errexit("UGetHostIP: gethostbyname(%s) failed!", hostname);

 return ntohl(inet_addr(inet_ntoa(*((struct in_addr *)host->h_addr))));
}


USOCKET UWaitAndAccept_Socket_IP(USOCKET sd, USocketOptions *sopt, UHOSTIP *ip)
{
 fd_set	                fdset;
 int                    sel;
 struct sockaddr_in     addr;
 USOCKET				newsd;
 USOCKLEN              len=sizeof(addr);
 
 
 FD_ZERO(&fdset);
 FD_SET(sd, &fdset);
 sel = select(sd+1, &fdset, NULL, NULL, NULL);
 if (!sel)
     errexit("UWaitAndAccept_Socket: select failed with sel=0");
     
 newsd = accept(sd, (struct sockaddr *)&addr, &len);
 if (newsd < 0)
     uprintf("UWaitAndAccept_Socket: Error in accepting connection\n");
 else
     _USetupSocket(newsd, sopt);

 *ip = ntohl(inet_addr(inet_ntoa(addr.sin_addr)));
 return newsd;
}


USOCKET UWaitAndAccept_Socket(USOCKET sd, USocketOptions *sopt)
{
 UHOSTIP ip;
 return UWaitAndAccept_Socket_IP(sd, sopt, &ip);
}


#if 0
    /* Converts ascii text to in_addr struct.  NULL is returned if the
       address can not be found. */
struct in_addr *atoaddr(char *address) {
         struct hostent *host;
         static struct in_addr saddr;

             /* First try it as aaa.bbb.ccc.ddd.  */
         saddr.s_addr = inet_addr(address);
         if (saddr.s_addr != -1) {
           return &saddr;
         }
         host = gethostbyname(address);
         if (host != NULL) {
           return (struct in_addr *) *host->h_addr_list;
         }
         return NULL;
       }
#endif



int URead_Raw(USOCKET sd, void *buf, int size)
{
  int i, j;
  char *b;

#if defined(_DEBUG)
  int size_in = size;
  printf("URead_Raw(sd=%d, buf=%x, size=%d) entry...\n", sd, (int)buf, size_in);
#endif 

  b = (char *)buf;
  i = size;
  while (i)
      {
      j = readsocket(sd, b, i); /* blocks until there is something <= i to return */
      b += j;
      i -= j;
      if (j <= 0)
          {
          size = -1;   /* Connection broken (or some other error). */
          break;
          }
      }
  
#if defined(_DEBUG)
  printf("URead_Raw(sd=%d, buf=%x, size=%d) exit rtn=%d.\n", sd, (int)buf, size_in, size);
#endif 
  return size;
}

int UWrite_Raw(USOCKET sd, const void *buf, int size)
{
  int i, j;
  const char *b;

#if defined(_DEBUG)
  int size_in = size;
  printf("UWrite_Raw(sd=%d, buf=%x, size=%d) entry...\n", sd, (int)buf, size_in);
#endif 
  b = (const char *)buf;
  i = size;
  while (i)
      {
      j = writesocket(sd, b, i); /* blocks until there is something <= i to return */
      b += j;
      i -= j;
      if (j <= 0)
          {
          size = -1;   /* Connection broken (or some other error). */
          break;
          }
      }
  
#if defined(_DEBUG)
  printf("UWrite_Raw(sd=%d, buf=%x, size=%d) exit rtn=%d.\n", sd, (int)buf, size_in, size);
#endif 
  return size;
}



int UWrite_Str(USOCKET sd, const char *st)
{
 int l=strlen(st);
 UWrite_Int(sd, &l, 1);
 return UWrite_Raw(sd, st, l);
}

char *URead_Str(USOCKET sd)
{
 int  l;
 char *st;
 URead_Int(sd, &l, 1);
 st = (char *) umalloc(l+1, "URead_Str: st");
 URead_Raw(sd, st, l);
 st[l] = 0;
 return st;
}


int UWrite_Int(USOCKET sd, int *ar, int n)
{
 int r;
 
 SwapByteArray(ar, sizeof(int), n);
 r = UWrite_Raw(sd, ar, n*sizeof(int)) / sizeof(int);
 SwapByteArray(ar, sizeof(int), n);
 return r;
}


int URead_Int(USOCKET sd, int *ar, int n)
{
 int r;
 
 r = URead_Raw(sd, ar, n*sizeof(int)) / sizeof(int);
 SwapByteArray(ar, sizeof(int), n);
 return r;
}



int UWrite_Long(USOCKET sd, long *ar, int n)
{
 int r;
 
 SwapByteArray(ar, sizeof(long), n);
 r = UWrite_Raw(sd, ar, n*sizeof(long)) / sizeof(long);
 SwapByteArray(ar, sizeof(long), n);
 return r;
}


int URead_Long(USOCKET sd, long *ar, int n)
{
 int r;
 
 r = URead_Raw(sd, ar, n*sizeof(long)) / sizeof(long);
 SwapByteArray(ar, sizeof(long), n);
 return r;
}


int UWrite_Float(USOCKET sd, float *ar, int n)
{
 int r;
 
 SwapByteArray(ar, sizeof(float), n);
 r = UWrite_Raw(sd, ar, n*sizeof(float)) / sizeof(float);
 SwapByteArray(ar, sizeof(float), n);
 return r;
}


int URead_Float(USOCKET sd, float *ar, int n)
{
 int r;
 
 r = URead_Raw(sd, ar, n*sizeof(float)) / sizeof(float);
 SwapByteArray(ar, sizeof(float), n);
 return r;
}


int UWrite_Double(USOCKET sd, double *ar, int n)
{
 int r;
 
 SwapByteArray(ar, sizeof(double), n);  
 r = UWrite_Raw(sd, ar, n*sizeof(double)) / sizeof(double);
 SwapByteArray(ar, sizeof(double), n);
 return r;
}


int URead_Double(USOCKET sd, double *ar, int n)
{
 int r;
 
 r = URead_Raw(sd, ar, n*sizeof(double)) / sizeof(double);
 SwapByteArray(ar, sizeof(double), n);
 return r;
}

