/*
--------------------------------------------------------------------
 This file is part of PaToH (Partitioning Tools for Hypergraphs)
 Copyright (c) 1996-   Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README file in the main directory.
---------------------------------------------------------------------
*/

// -*- c++ -*-
// ****************************************************************************
//  Name: uarray.h
//  Athr: Umit V. Catalyurek (UVC)
//  Date: 09/29/2001 : UVC - created based on Mike Beynon's DArray
//
//  Desc: Implementation of a dynamic array class.
// ****************************************************************************

#if !defined(_UUARRAY_H_)
#define _UUARRAY_H_

#include <iostream>
#include "usocket.h"

using namespace std;

typedef unsigned int uint;

//
// Notes: T must support assignment and have a constructor with no args
//
template <class T>
class UUArray
{
private:
#define DefaultExtentSize  10
protected:
  T*    lst;
  uint maxsize;
  uint size;
  uint extentsize;

    void extendArray(uint nExtent=0)
        {
        if (!nExtent)
            nExtent = extentsize;
        
        uint nNewSize = maxsize + nExtent;
        T* rgTmp = lst;

//        uprintf("newsize=%u\n", nNewSize);
        
        lst = new T[maxsize = nNewSize];
        if (rgTmp)
            {
                // copy over old elements
            for (uint i=0; i<size; i++) 
                lst[i] = rgTmp[i];
            delete [] rgTmp;
            }
        }

public:
    UUArray(uint maxsize_in = 0, uint extentsize_in = DefaultExtentSize)
        {
        maxsize = maxsize_in;
        if (maxsize == 0) 
            lst = NULL;
        else 
            lst = new T[maxsize];
        size = 0;
        extentsize = extentsize_in;
        }
    UUArray(const UUArray<T>& darray)
        {
        size = 0;
        lst = NUL;
        *this = darray;
        }
    
    ~UUArray()
        {
        if (lst)
            delete [] lst;
        }
    
    uint getSize(void)                        { return size; }
    void setExtentSize(uint extentsize_in)   { extentsize = extentsize_in; }

    T& operator[](uint i)
        {
        if (i >= maxsize)
            extendTo(i);
        if (i >= size)
            size = i+1;
        return lst[i];
        }

    const T& operator[](uint i) const          { return lst[i]; }

    void extendTo(int newmax)
        {
        extendArray(newmax-maxsize+extentsize);
        }
    
    
    uint append(const T& elem)
        {
//        cout << "append(const T& elem)" << elem << endl;
        if (size == maxsize)
            extendArray();
        lst[size++] = elem;
        return size-1;  // index of where added
        }

    uint appendByVal(const T elem)
        {
//        cout << "appendByVal(const T elem)" << elem << endl;
        return append(elem);
        }
    
            // remove an element at loc 'i'
    uint remove(uint i)
        {
        if (i >= size)
            return size;  // don't do anything
            // copy all the elements
        for (uint j=i+1; j<size; ++j) 
            lst[j-1] = lst[j];
        return --size;
        }

    void clear()
        {
        if (lst)
            delete[] lst;
        
        lst = NULL;
        maxsize = size = 0;
        }


        // whole array assignment
    UUArray<T>& operator=(const UUArray<T>& darray)
        {
        if (this != &darray)
            {  // avoid self copies
            if (maxsize > 0)
                delete [] lst;
            maxsize = size = darray.size;
            if (maxsize > 0) 
                lst = new T[maxsize];        
            for (uint i=0; i<size; i++) 
                lst[i] = darray.lst[i];
            }
        return *this;
        }

    friend std::ostream& operator<<(ostream& os, UUArray<T>& darray)
        {
        for (uint i=0; i<darray.size; i++)
            os << "[" << i << "] " << darray.lst[i] << endl;
        return os;
        }
};

class UUIntArray : public UUArray<int>
{
public:
    UUIntArray() :  UUArray<int>(0)
        {}
    
    UUIntArray(char *readst)
        {
        read(readst);
        }

    uint isIn(int v)
        {
        uint i;
        
        for (i=0; i<getSize() && lst[i]!=v; ++i);
        if (i>=getSize())
            return 0;
        return i;
        }    

    int read(char *st)
        {
        char *p;

        for (p=st; *p; ++p)
            if (*p==',')
                *p = ' ';        
        
        uint sz=(uint)strtol(st, &p, 10);
//        uprintf("read from string sz=%u [%s/%s]\n", sz, st, p);
        clear();
        if (sz)
            {
            extendTo(sz);
            for (uint i=0; i<sz && p; ++i)
                {
                st = p;
                appendByVal((int)strtol(st, &p, 10));
                }            
            }
        return sz;
        }
    
    int write(char *st, int len)
        {
        char s[20];
        int  tl, l;
        
        sprintf(st, "%d,", getSize());
        tl = 1 + strlen(st);
        for (uint i=0; i<getSize(); ++i)
            {
            sprintf(s, "%d,", lst[i]);
            l = strlen(s);
            if (tl+l<len)
                {
                strcat(st, s);
                tl += l;
                }
            else
                return -1;
            }
        return getSize();
        }


    int read(int sd)
        {
        uint sz;
        
        URead_Int(sd, (int *) &sz, 1);
//        uprintf(" read from socket sz=%u :[", sz);
        clear();
        if (sz)
            {
            extendTo(sz);
            for (uint i=0; i<sz; ++i)
                {
                int v;
                URead_Int(sd, &v, 1);
//                uprintf("%d, ", v);
                appendByVal(v);
                }            
            }
        return sz;
        }
    
    int write(int sd)
        {
        int v=(int) getSize();
        
        UWrite_Int(sd, &v, 1);
        for (uint i=0; i<getSize(); ++i)
            UWrite_Int(sd, &lst[i], 1);
        return getSize();
        }

};



#endif 

