#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/stat.h>

#include "option.h"
#include "utils.h"
#include "dgraphReader.h"
#include "vcycle.h"
#include "dgraphPartition.h"


int processArgs_MLGP(int argc, char **argv, MLGP_option* opt)
{
    if (argc < 3) {
        printMLGPusage(argv[0]);
        u_errexit("%s: There is a problem with the arguments.\n", argv[0]);
    }

    initMLGPoptions(opt, atoi(argv[2]));
    processMLGPargs(argc, argv, opt);

    return 0;
}


void run_MLGP(char* file_name, MLGP_option opt)
{
    dgraph G;
    int i;
    int maxindegree, minindegree, maxoutdegree, minoutdegree;
    double aveindegree, aveoutdegree;

    readDGraph(&G, file_name, opt.use_binary_input);


    set_dgraph_info(&G);
    dgraph_info(&G, &maxindegree, &minindegree, &aveindegree, &maxoutdegree, &minoutdegree, &aveoutdegree);

    double ratio = opt.ratio;
    double eca = ((ratio-1.)*G.totvw)/opt.nbPart;

    for (i = 0; i<opt.nbPart; i++) {
        if (opt.lb[i] < 0) {
            opt.lb[i] = 1;
        }
        if (opt.ub[i] < 0) {
           opt.ub[i] = G.totvw/(opt.nbPart*1.0) + eca;
        }

        if ((floor(opt.lb[i]) < opt.lb[i])&&(floor(opt.lb[i]) == floor(opt.ub[i]))) {
            printf("WARNING: The balance can not be matched so we loosen it!!!\n");
            opt.lb[i] = floor(opt.lb[i]);
            opt.ub[i] = floor(opt.ub[i])+1;
        }
    }

    if (opt.debug)
        printf("Nb node: %d\nNb Edges: %d\nNb part: %d\nLower Bound: %f\nUper Bound: %f\n\n", G.nVrtx, G.nEdge, opt.nbPart, opt.lb[0], opt.ub[0]);

    ecType* edgecut = (ecType*) malloc(sizeof(ecType) * opt.runs);
    ecType* nbcomm = (ecType*) malloc(sizeof(ecType) * opt.runs);
    double* latencies = (double*) malloc(sizeof(double) * opt.runs);
    int r;
    coarsen* C=NULL;
    FILE * pFile = fopen(opt.out_file, "w");
    if( pFile == NULL)
        u_errexit(" run_MLGP: cannot open file %s\n", opt.out_file);
    char *out_filep = opt.file_name;
    out_filep[strlen(out_filep)-4]='\0';

    usRandom(opt.seed);

    for (r = 0; r<opt.runs; r++) {
        printf("########################## RUN %d (seed=%d) ########################\n", r, uGetSeed());
        MLGP_info* info = (MLGP_info*) malloc (sizeof (MLGP_info));
        initInfoPart(info);

        idxType* flag = (idxType * ) calloc(G.nVrtx +1, sizeof(idxType));
        C = VCycle(&G, flag, opt, info);
        free(flag);

        edgecut[r] = edgeCut(&G, C->part);
        latencies[r] = computeLatency(&G, C->part, 1.0, 11.0);
        printf("Final run %d\t%d\t%d\t%d\t%d\t%f\n", r, G.nVrtx, G.nEdge, (int) edgecut[r], (int) nbcomm[r], latencies[r]);
        printf("Times\t%.3lf\t%.3lf\t%.3lf\t%.3lf\n", info->timing_coars, info->timing_inipart, info->timing_uncoars,info->timing_global);
        printf("depth %d\tRel nVrtx last %f\tRel nEdge last %f\n", info->coars_depth, (double) info->nbnodes_coars_tab[info->coars_depth] / (double) G.nVrtx, (double) info->nbedges_coars_tab[info->coars_depth] / (double) G.nEdge);
        printf("######################## END RUN %d ########################\n\n", r);
        if (opt.print > 0) {
            printf("######################## OUTPUT %d ########################\n\n", r);
            printInfoPart(info, C->graph->nVrtx, C->graph->nEdge, opt.print);
            printf("\n###########################################################\n\n");
        }
        freeInfoPart(info);
    }
    fclose (pFile);

    double edgecutave = 0.0, nbcommave = 0.0, edgecutsd = 0.0, nbcommsd = 0.0, latencyave = 0.0;
    for (r = 0; r<opt.runs; r++) {
        edgecutave += edgecut[r];
        nbcommave += nbcomm[r];
        latencyave += latencies[r];
    }
    edgecutave = edgecutave / opt.runs;
    nbcommave = nbcommave / opt.runs;
    latencyave = latencyave / opt.runs;
    for (r = 0; r<opt.runs; r++) {
        edgecutsd = edgecut[r] - edgecutave < 0? edgecutave - edgecut[r] : edgecut[r] - edgecutave;
        nbcommsd = nbcomm[r] - nbcommave < 0? nbcommave - nbcomm[r] : nbcomm[r] - nbcommave;
    }
    edgecutsd = edgecutsd / opt.runs;
    nbcommsd = nbcommsd / opt.runs;
    printf("Final runs\t%d\t%d\t%f\t%f\t%f\t%f\t%f\n", G.nVrtx, G.nEdge, edgecutave, edgecutsd, nbcommave, nbcommsd, latencyave);

    freeDGraphData(&G);
    free(edgecut);
    free(nbcomm);
    free(latencies);
    freeCoarsenHeader(C);
}


int main(int argc, char *argv[])
{
    MLGP_option opt;
    processArgs_MLGP(argc, argv, &opt);

    run_MLGP(opt.file_name, opt);

    free_opt(&opt);

    return 0;
}
