#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>

#include "option.h"
#include "vcycle.h"
#include "dgraph.h"
#include "utils.h"
#include "info.h"
#include "dgraphPartition.h"
#include "debug.h"
#include "vcycle.h"
#include "refinement.h"


static inline idxType computeMovabletoUpNode(dgraph* G, idxType* part, int nbpart, idxType* topsortpart, idxType* toporderpart, idxType* partsize, idxType* movableto, idxType i) {
    idxType moveup = -1, j;
    for (j = G->inStart[i]; j <= G->inEnd[i]; j++) {
        idxType innode = G->in[j];
        idxType parti = part[innode];
        if (moveup == -1 && (topsortpart[parti] <= topsortpart[part[i]]))
            moveup = parti;
        if ((topsortpart[parti] > topsortpart[moveup]) && (topsortpart[parti] <= topsortpart[part[i]]))
            moveup = parti;
    }
    if (moveup == -1)
        moveup = topsortpart[part[i]] == 0 ? part[i] : toporderpart[topsortpart[part[i]] - 1];
    return moveup;
}

static inline idxType computeMovabletoDownNode(dgraph* G, idxType* part, int nbpart, idxType* topsortpart, idxType* toporderpart, idxType* partsize, idxType* movableto, idxType i) {
    idxType movedown = -1, j;
    for (j = G->outStart[i]; j <= G->outEnd[i]; j++) {
        idxType outnode = G->out[j];
        idxType parti = part[outnode];
        if (movedown == -1 && (topsortpart[parti] >= topsortpart[part[i]]))
            movedown = parti;
        if ((topsortpart[parti] < topsortpart[movedown]) && (topsortpart[parti] >= topsortpart[part[i]]))
            movedown = parti;
    }
    if (movedown == -1)
        movedown = topsortpart[part[i]] == nbpart - 1 ? part[i] : toporderpart[topsortpart[part[i]] + 1];
    return movedown;
}

static inline void computeMovabletoNode(dgraph* G, idxType* part, int nbpart, idxType* topsortpart, idxType* toporderpart, idxType* partsize, idxType* movableto, idxType i) {
    idxType moveup = computeMovabletoUpNode(G, part, nbpart, topsortpart, toporderpart, partsize, movableto, i);
    idxType movedown = computeMovabletoDownNode(G, part, nbpart, topsortpart, toporderpart, partsize, movableto, i);
    if ((movedown == part[i])&&(moveup == part[i]))
		movableto[i] = part[i];
    else if ((movedown != part[i])&&(moveup != part[i]))
		if (partsize[movedown] < partsize[moveup])
		    movableto[i] = movedown;
		else
		    movableto[i] = moveup;
    else if (moveup != part[i])
		movableto[i] = moveup;
    else
		movableto[i] = movedown;
}


inline ecType computeGainNode(dgraph *G, idxType *part, idxType movableto, idxType node, MLGP_option opt)
{
    idxType j;
    ecType newgain = 0;

    switch (opt.obj){
        case CO_OBJ_EC :
            for (j=G->inStart[node]; j<=G->inEnd[node]; j++){
                idxType innode = G->in[j];
                if (part[innode] == part[node])
                    newgain -= G->ecIn[j];
                else if (part[innode] == movableto)
                    newgain += G->ecIn[j];
            }
            for (j=G->outStart[node]; j<=G->outEnd[node]; j++){
                idxType outnode = G->out[j];
                if (part[outnode] == part[node])
                    newgain -= G->ecOut[j];
                else if (part[outnode] == movableto)
                    newgain += G->ecOut[j];
            }
            break;
        case CO_OBJ_CV :
            for (j=G->inStart[node]; j<=G->inEnd[node]; j++){
                idxType innode = G->in[j];
                if (part[innode] == part[node])
                    newgain -= G->ecIn[j];
                else if (part[innode] == movableto)
                    newgain += G->ecIn[j];
            }
            for (j=G->outStart[node]; j<=G->outEnd[node]; j++){
                idxType outnode = G->out[j];
                if (part[outnode] == part[node])
                    newgain -= G->ecOut[j];
                else if (part[outnode] == movableto)
                    newgain += G->ecOut[j];
            }
            break;
    }
    return newgain;
}


static inline void kwayInitializeGainAndHeap(dgraph *G, ecType* gain, idxType* movableto, idxType* heap, idxType* inheap, idxType* hsize, idxType nbpart, idxType *part, idxType* partsize, idxType* maxpartsize, idxType* toporderpart, idxType* topsortpart, idxType* partcopy, MLGP_option opt) {
    /* This function is called at the begining of refinements for kway.
     * It initialize the gain table, the movableto table and build the heap
     * It also initializes topsortpart, partsize, maxpartsize and partcopy
    */
    idxType i;
    for (i=0; i<nbpart; i++){
        topsortpart[toporderpart[i]] = i;
    }
    for (i=1; i<=G->nVrtx; i++){
        if (partcopy != NULL)
            partcopy[i] = part[i];
        computeMovabletoNode(G, part, nbpart, topsortpart, toporderpart, partsize, movableto, i);
        gain[i] = computeGainNode(G, part, movableto[i], i, opt);
        if (movableto[i] != part[i]){
            (*hsize)++;
            heap[*hsize] = i;
            inheap[i] = *hsize;
        }
        partsize[part[i]] += G->vw[i];
        *maxpartsize = *maxpartsize < partsize[part[i]] ? partsize[part[i]] : *maxpartsize;
    }

    heapBuild(heap, gain, *hsize, inheap);
}



static inline void kwayMoveAndUpdate(dgraph *G, idxType nodeToMove, idxType* movableto, idxType* moved, idxType* nbmoved, ecType* edgecut, ecType* gain, idxType *part, idxType nbpart, idxType* partsize, idxType* maxpartsize, idxType* toporderpart, idxType* topsortpart, idxType* heap, idxType* hsize, idxType* inheap, int* anchored, MLGP_option opt) {
    /* This function is called in the different k-way refinements.
     * It moves nodeToMove from part[nodeToMove] to movableto[nodeToMove]
     * Update the partsize, the heap, anchored table, moved table, maxpartsize value and the edgecut
     */
    idxType i,j;
    idxType origin_part = part[nodeToMove];
    idxType dest_part = movableto[nodeToMove];

    part[nodeToMove] = dest_part;
    partsize[origin_part] -= G->vw[nodeToMove];
    partsize[dest_part] += G->vw[nodeToMove];
    anchored[nodeToMove] = 1;
    moved[*nbmoved] = nodeToMove;
    (*nbmoved)++;
    *edgecut -= gain[nodeToMove];

    *maxpartsize = 0;
    for (i=0; i<nbpart; i++)
        *maxpartsize = *maxpartsize < partsize[i] ? partsize[i] : *maxpartsize;

    for (j=G->inStart[nodeToMove]; j<=G->inEnd[nodeToMove]; j++){
        idxType father = G->in[j];
        if(anchored[father]) continue; /*we do not even update their connectivity*/
        computeMovabletoNode(G, part, nbpart, topsortpart, toporderpart, partsize, movableto, father);
        gain[father] = computeGainNode(G, part, movableto[father], father, opt);
        if (inheap[father] > 0)
            continue;
        if (movableto[father] != part[father]){
            heapInsert(heap, gain, hsize, father, inheap);
        }
    }

    for (j=G->outStart[nodeToMove]; j<=G->outEnd[nodeToMove]; j++){
        idxType child = G->out[j];
        if(anchored[child]) continue; /*we do not even update their connectivity*/
        computeMovabletoNode(G, part, nbpart, topsortpart, toporderpart, partsize, movableto, child);
        gain[child] = computeGainNode(G, part, movableto[child], child, opt);
        if (inheap[child] > 0)
            continue;
        if (movableto[child] != part[child]){
            heapInsert(heap, gain, hsize, child, inheap);
        }
    }
}


int createCycle(dgraph* G, idxType* part, idxType nbpart, ecType** partmatrix, idxType node, idxType movableto)
{
    idxType part_save = part[node];
    int flag = computePartmatrix(G, part, partmatrix, node, movableto);
    part[node] = movableto;
    int res = 0;
    if (flag)
        res = thereIsCycle(nbpart, partmatrix);
    computePartmatrix(G, part, partmatrix, node, part_save);
    part[node] = part_save;
    return res;
}


static inline void computeMovabletoBestPartNode(dgraph* G, idxType* part, idxType* partsize, int nbpart, idxType i, idxType* movableto, ecType* gain, MLGP_option opt)
{
    idxType j,k,idx = 0;

    for (j=0; j<nbpart; j++){
    	if (j==part[i])
    	    continue;
    	movableto[idx] = j;
    	gain[idx] = computeGainNode(G, part, j, i, opt);
    	for (k=idx; k>=1; k--){
    	    if (gain[k] > gain[k-1]){
    		    idxType movable_save = movableto[k];
    		    idxType gain_save = gain[k];
    		    movableto[k] = movableto[k-1];
    		    gain[k] = gain[k-1];
    		    movableto[k-1] = movable_save;
    		    gain[k-1] = gain_save;
    	    }
    	}
    	idx++;
    }
}

static inline int isIncomingBoundary(dgraph* G, idxType* part, idxType node)
{
    idxType i;
    for (i=G->inStart[node]; i<=G->inEnd[node]; i++)
	    if (part[G->in[i]] == part[node])
	        return 0;
    return 1;
}

static inline int isOutgoingBoundary(dgraph* G, idxType* part, idxType node)
{
    idxType i;
    for (i=G->outStart[node]; i<=G->outEnd[node]; i++)
	    if (part[G->out[i]] == part[node])
	        return 0;
    return 1;
}

void forcedBalance(dgraph *G, idxType *toporderpart,
		   double* ub_pw,
		      idxType *part, idxType nbpart, MLGP_option opt, MLGP_info* info, idxType* partsize)
{
    idxType i, hsize = 0, maxpartsize = 0;
    idxType nbmoved = 0,  newsize1, newsize2, node = 1;
    idxType* movableto = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* nbmovableto = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* topsortpart = (idxType*) calloc (nbpart, sizeof(idxType));
    idxType* moved = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    ecType* gain = (ecType*) malloc(sizeof(ecType)*(G->nVrtx+1));
    idxType* heap = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));
    idxType* inheap = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    int* anchored = (int*) calloc(G->nVrtx+1, sizeof(idxType));

    for (i=0; i<nbpart; i++)
		partsize[i] = 0;

    kwayInitializeGainAndHeap(G, gain, movableto, heap, inheap, &hsize, nbpart, part, partsize, &maxpartsize, toporderpart, topsortpart, NULL, opt);

    while ((partSizeChecker(partsize, ub_pw, opt.nbPart, G->maxVW))&&(hsize > 0)){
	    node = heapExtractMax(heap, gain, &hsize, inheap);

	    newsize1 = partsize[part[node]] - G->vw[node];
	    newsize2 = partsize[movableto[node]] + G->vw[node];

	    if (newsize2 > ub_pw[movableto[node]])
	        continue;

       kwayMoveAndUpdate(G, node, movableto, moved, &nbmoved, &(info->current_edgecut), gain, part, nbpart, partsize, &maxpartsize, toporderpart, topsortpart, heap, &hsize, inheap, anchored, opt);
    }
    if(opt.debug > 3)
    	printf("moved weight %d\n",nbmoved );

    free(movableto);
    free(nbmovableto);
    free(topsortpart);
    free(moved);
    free(gain);
    free(heap);
    free(inheap);
    free(anchored);
}



void refinementPostOrder(dgraph *G, idxType *toporderpart,
                         double* lb_pw, double* ub_pw,
                         idxType *part, idxType nbpart, ecType* edgecut, MLGP_option opt)
{
    idxType i, hsize = 0, maxpartsize = 0, nbmoved = 0, newsize1, newsize2, node = 1;
    idxType without_impro = 0, minindex = 0;
    idxType* movableto = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* partsize = (idxType*) calloc(nbpart+1, sizeof(idxType));
    idxType* partcopy = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* topsortpart = (idxType*) malloc (sizeof(idxType) * nbpart);
    idxType* moved = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    ecType* gain = (ecType*) malloc(sizeof(ecType)*(G->nVrtx+1));
    idxType* heap = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* inheap = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));
    idxType* anchored = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));

    kwayInitializeGainAndHeap(G, gain, movableto, heap, inheap, &hsize, nbpart, part, partsize, &maxpartsize, toporderpart, topsortpart, partcopy, opt);

    idxType minmaxpartsize = maxpartsize;
    ecType minedgecut = *edgecut;

    while ((hsize > 0)&&(without_impro <= G->nVrtx/4)&&(nbmoved<G->nVrtx)){
        node = heapExtractMax(heap, gain, &hsize, inheap);

        newsize1 = partsize[part[node]] - G->vw[node];
        newsize2 = partsize[movableto[node]] + G->vw[node];
        if ((newsize1 < lb_pw[part[node]])||(newsize1 > ub_pw[part[node]])||(newsize2 < lb_pw[movableto[node]])||(newsize2 > ub_pw[movableto[node]]))
            continue;

        kwayMoveAndUpdate(G, node, movableto, moved, &nbmoved, edgecut, gain, part, nbpart, partsize, &maxpartsize, toporderpart, topsortpart, heap, &hsize, inheap, anchored, opt);

        if ((*edgecut < minedgecut)||((*edgecut == minedgecut)&&(maxpartsize <= minmaxpartsize))){
            minedgecut = *edgecut;
            minmaxpartsize = maxpartsize;
            minindex = nbmoved;
            without_impro = -1;
        }
        without_impro ++;
    }
    for (i = nbmoved-1; i >= minindex; i--){
        part[moved[i]] = partcopy[moved[i]];
    }

    *edgecut = minedgecut;
    free(movableto);
    free(partsize);
    free(partcopy);
    free(topsortpart);
    free(moved);
    free(gain);
    free(heap);
    free(inheap);
    free(anchored);
}

void refinementRandomizedSelectTowardsBest(dgraph *G, ecType** partmatrix,
              double* lb_pw, double* ub_pw,
              idxType *part, idxType nbpart, ecType* edgecut, MLGP_option opt)
{
    idxType i,j,k, qsize = 0;
    idxType maxpartsize = 0;
    ecType** gain = (ecType**) malloc(sizeof(ecType*)*(G->nVrtx+1));
    idxType* queue = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    int* inqueue = (int*) calloc(G->nVrtx+1, sizeof(int));
    int* hasbeenmoved = (int*) calloc(G->nVrtx+1, sizeof(int));
    idxType* moved = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType** movableto = (idxType**) malloc(sizeof(idxType*)*(G->nVrtx+1));
    idxType* partsize = (idxType*) calloc(nbpart+1, sizeof(idxType));
    idxType* partcopy = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));

    for (i = 1; i <= G->nVrtx; i++){
        partcopy[i] = part[i];
        partsize[part[i]] += G->vw[i];
        movableto[i] = (idxType*) umalloc(sizeof(idxType) * (nbpart-1),"movableto[i]");
        gain[i] = (ecType*) umalloc(sizeof(ecType) * (nbpart-1),"gain[i]");
    }

    for (i = 1; i <= G->nVrtx; i++){
        if ((isIncomingBoundary(G, part, i))||(isOutgoingBoundary(G, part, i))){
            queue[qsize++] = i;
            inqueue[i] = 1;
        }
    }

    for (i = 0; i < nbpart; i++){
        maxpartsize = maxpartsize < partsize[i] ? partsize[i] : maxpartsize;
    }

    idxType nbmoved = 0, newsize1, newsize2, node = 1, minmaxpartsize = maxpartsize;
    int without_impro = 0, minindex = 0;
    ecType minedgecut = *edgecut;
    if(qsize>0)
        shuffleArray(qsize, queue, qsize/4+1);
    while ((qsize > 0)&&(without_impro <= G->nVrtx/4)&&(nbmoved<G->nVrtx)){

        node = queue[qsize-1];
        --qsize;
        inqueue[node] = 0;
        computeMovabletoBestPartNode(G, part, partsize, nbpart, node, movableto[node], gain[node], opt);

        if ((movableto[node][0] == -1)||(hasbeenmoved[node]))
            continue;

        if(gain[node][0]<0)
            continue;

        newsize1 = partsize[part[node]] - G->vw[node];
        newsize2 = partsize[movableto[node][0]] + G->vw[node];
        int cycle = createCycle(G, part, nbpart, partmatrix, node, movableto[node][0]);

        k=0;
        while (((newsize1 < lb_pw[part[node]])||(newsize1 > ub_pw[part[node]])||(newsize2 < lb_pw[movableto[node][k]])||(newsize2 > ub_pw[movableto[node][k]])||(cycle))&&(k<nbpart-2)){
            //Let's try second best partition for this node
            k++;
            if (gain[node][k] < gain[node][1])
                break;
            if(gain[node][k]<0)
                continue;
            newsize2 = partsize[movableto[node][k]] + G->vw[node];
            cycle = createCycle(G, part, nbpart, partmatrix, node, movableto[node][k]);
        }
        if ((newsize1 < lb_pw[part[node]])||(newsize1 > ub_pw[part[node]])||(newsize2 < lb_pw[movableto[node][k]])||(newsize2 > ub_pw[movableto[node][k]])||(cycle)||(k>=nbpart)||(gain[node][k] < gain[node][1]))
            continue;

        if(qsize>0)
            shuffleArray(qsize, queue, qsize/4+1);
        computePartmatrix(G, part, partmatrix, node, movableto[node][k]);

        partsize[part[node]] = newsize1;
        partsize[movableto[node][k]] = newsize2;
        maxpartsize = 0;
        for (j=0; j<nbpart; j++)
            maxpartsize = maxpartsize < partsize[j] ? partsize[j] : maxpartsize;
        *edgecut -= gain[node][k];
        part[node] = movableto[node][k];
        hasbeenmoved[node] = 1;
        moved[nbmoved++] = node;

        for (j=G->inStart[node]; j<=G->inEnd[node]; j++){
            idxType father = G->in[j];
            if (!((isIncomingBoundary(G, part, father))||(isOutgoingBoundary(G, part, father))))
                continue;
            if ((inqueue[father] == 0)&&(hasbeenmoved[father] == 0)){
                queue[qsize++] = father;
                inqueue[father] = 1;
            }
        }
        for (j=G->outStart[node]; j<=G->outEnd[node]; j++){
            idxType child = G->out[j];
            if (!((isIncomingBoundary(G, part, child))||(isOutgoingBoundary(G, part, child))))
                continue;
            if ((inqueue[child] == 0)&&(hasbeenmoved[child] == 0)){
                queue[qsize++] = child;
                inqueue[child] = 1;
            }
        }

        if ((*edgecut < minedgecut)||((*edgecut == minedgecut)&&(maxpartsize <= minmaxpartsize))){
            minedgecut = *edgecut;
            minmaxpartsize = maxpartsize;
            minindex = nbmoved;
            without_impro = -1;
        }
        without_impro ++;

    }

    for (i = nbmoved-1; i >= minindex; i--){
        part[moved[i]] = partcopy[moved[i]];
    }

    *edgecut = minedgecut;
    for (i = 1; i <= G->nVrtx; i++){
        free(gain[i]);
        free(movableto[i]);
    }
    free(gain);
    free(queue);
    free(inqueue);
    free(hasbeenmoved);
    free(moved);
    free(movableto);
    free(partsize);
    free(partcopy);
}

void refinementTowardBest(dgraph *G, ecType** partmatrix,
			  double* lb_pw, double* ub_pw,
			  idxType *part, idxType nbpart, ecType* edgecut, MLGP_option opt)
{
    idxType i,j,k, hsize = 0;
    idxType maxpartsize = 0;
    ecType** gain = (ecType**) malloc(sizeof(ecType*)*(G->nVrtx+1));
    ecType* gain_0 = (ecType*) malloc(sizeof(ecType)*(G->nVrtx+1));
    idxType* heap = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType* inheap = (idxType*) calloc(G->nVrtx+1, sizeof(idxType));
    int* hasbeenmoved = (int*) calloc(G->nVrtx+1, sizeof(int));
    idxType* moved = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));
    idxType** movableto = (idxType**) malloc(sizeof(idxType*)*(G->nVrtx+1));
    idxType* partsize = (idxType*) calloc(nbpart+1, sizeof(idxType));
    idxType* partcopy = (idxType*) malloc(sizeof(idxType)*(G->nVrtx+1));

    for (i = 1; i <= G->nVrtx; i++){
    	partsize[part[i]] += G->vw[i];
    	partcopy[i] = part[i];
    	movableto[i] = (idxType*) malloc(sizeof(idxType) * (nbpart-1));
    	gain[i] = (ecType*) malloc(sizeof(ecType) * (nbpart-1));
    }
    for (i = 1; i <= G->nVrtx; i++){
    	if ((isIncomingBoundary(G, part, i))||(isOutgoingBoundary(G, part, i))){
    	    computeMovabletoBestPartNode(G, part, partsize, nbpart, i, movableto[i], gain[i], opt);
    	    gain_0[i] = gain[i][0];
    	    heap[++hsize] = i;
    	    inheap[i] = hsize;
    	}
    }
    heapBuild(heap, gain_0, hsize, inheap);
    for (i = 0; i < nbpart; i++){
	    maxpartsize = maxpartsize < partsize[i] ? partsize[i] : maxpartsize;
    }

    idxType nbmoved = 0, newsize1, newsize2, node = 1, minmaxpartsize = maxpartsize;
    int without_impro = 0, minindex = 0;
    ecType minedgecut = *edgecut;

    while ((hsize > 0)&&(without_impro <= G->nVrtx/4)&&(nbmoved<G->nVrtx)){
    	node = heapExtractMax( heap, gain_0, &hsize, inheap);

    	if ((movableto[node][0] == -1)||(hasbeenmoved[node]))
    	    continue;
    	newsize1 = partsize[part[node]] - G->vw[node];
    	newsize2 = partsize[movableto[node][0]] + G->vw[node];

    	int cycle = createCycle(G, part, nbpart, partmatrix, node, movableto[node][0]);
    	k=0;
    	while (((newsize1 < lb_pw[part[node]])||(newsize1 > ub_pw[part[node]])||(newsize2 < lb_pw[movableto[node][k]])||(newsize2 > ub_pw[movableto[node][k]])||(cycle))&&(k<nbpart-2)){
    	    //Let's try second best partition for this node
    	    k++;
    	    if (gain[node][k] < gain[node][1])
    		break;
    	    newsize2 = partsize[movableto[node][k]] + G->vw[node];
    	    cycle = createCycle(G, part, nbpart, partmatrix, node, movableto[node][k]);
    	}
    	if ((newsize1 < lb_pw[part[node]])||(newsize1 > ub_pw[part[node]])||(newsize2 < lb_pw[movableto[node][k]])||(newsize2 > ub_pw[movableto[node][k]])||(cycle)||(k>=nbpart)||(gain[node][k] < gain[node][1]))
    	    continue;

    	computePartmatrix(G, part, partmatrix, node, movableto[node][k]);

    	partsize[part[node]] = newsize1;
    	partsize[movableto[node][k]] = newsize2;
    	maxpartsize = 0;
    	for (j=0; j<nbpart; j++)
    	    maxpartsize = maxpartsize < partsize[j] ? partsize[j] : maxpartsize;
    	*edgecut -= gain[node][k];
    	part[node] = movableto[node][k];
    	hasbeenmoved[node] = 1;
    	moved[nbmoved++] = node;

    	for (j=G->inStart[node]; j<=G->inEnd[node]; j++){
    	    idxType father = G->in[j];
    	    if (!((isIncomingBoundary(G, part, father))||(isOutgoingBoundary(G, part, father))))
    		    continue;
    	    computeMovabletoBestPartNode(G, part, partsize, nbpart, father, movableto[father], gain[father], opt);
    	    ecType newgain = gain[father][0];
    	    if ((inheap[father] == 0)&&(hasbeenmoved[father] == 0)){
                gain_0[father] = newgain;
                heapInsert(heap, gain_0, &hsize, father, inheap);
    	    }
            if ((inheap[father] != 0)&&(hasbeenmoved[father] == 0)){
                heapUpdateKey(heap, inheap, gain_0, hsize, father, newgain);
            }
    	}
    	for (j=G->outStart[node]; j<=G->outEnd[node]; j++){
    	    idxType child = G->out[j];
    	    if (!((isIncomingBoundary(G, part, child))||(isOutgoingBoundary(G, part, child))))
    		    continue;
    	    computeMovabletoBestPartNode(G, part, partsize, nbpart, child, movableto[child], gain[child], opt);
    	    ecType newgain = gain[child][0];
    	    if ((inheap[child] == 0)&&(hasbeenmoved[child] == 0)){
                gain_0[child] = newgain;
                heapInsert(heap, gain_0, &hsize, child, inheap);
    	    }
            if ((inheap[child] != 0)&&(hasbeenmoved[child] == 0)){
                heapUpdateKey(heap, inheap, gain_0, hsize, child, newgain);
            }
    	}

    	if ((*edgecut < minedgecut)||((*edgecut == minedgecut)&&(maxpartsize <= minmaxpartsize))){
    	    minedgecut = *edgecut;
    	    minmaxpartsize = maxpartsize;
    	    minindex = nbmoved;
    	    without_impro = -1;
    	}
    	without_impro ++;
    }

    for (i = nbmoved-1; i >= minindex; i--){
	part[moved[i]] = partcopy[moved[i]];
    }

    *edgecut = minedgecut;
    for (i = 1; i <= G->nVrtx; i++){
	free(gain[i]);
	free(movableto[i]);
    }
    free(gain);
    free(gain_0);
    free(heap);
    free(inheap);
    free(hasbeenmoved);
    free(moved);
    free(movableto);
    free(partsize);
    free(partcopy);
}


void KwayRefinementStep(MLGP_option opt, int level, dgraph* graph, idxType* part, idxType nbpart, MLGP_info* info)
{
    if (opt.refinement == REF_NONE)
        return;

    idxType j, i, maxpartsize = 0, previous_maxpartsize;
    vwType* partsize = (vwType*) calloc(opt.nbPart+1, sizeof(vwType));
    int nbbalance = 1;
    double sum = 0;
    double* bound = (double*) malloc(sizeof(double) * nbpart);
    idxType* toporder = (idxType*) malloc(sizeof(idxType) * nbpart);
    ecType lastobjvalue, objvalue = -1;
    ecType** partmatrix=NULL;

    switch(opt.refinement) {
        case REF_kFM_TB :
        case REF_kFM_POTB :
        case REF_KL :
        case REF_KL_bFM_MAXW :
        case REF_kFM_RANDV:
            partmatrix = (ecType **) malloc(opt.nbPart * sizeof(ecType * ));
            for (j = 0; j < opt.nbPart; j++)
                partmatrix[j] = (ecType *) malloc(opt.nbPart * sizeof(ecType));
            break;
    }

    for (i=1; i<=graph->nVrtx; i++)
        partsize[part[i]] += graph->vw[i];

    for (i = 0; i < nbpart; i++) {
        sum += opt.ub[i];
    }

    for (i = 0; i < nbpart; i++) {
        bound[i] = graph->totvw * opt.ub[i] / sum + 0.9 * (opt.ub[i] - graph->totvw * opt.ub[i] / sum);
        maxpartsize = maxpartsize < partsize[i] ? partsize[i] : maxpartsize;
    }

    info->timing_forced_tab[level] -= u_wseconds();

    randTopSortOnParts(graph, part, toporder, nbpart);

    objvalue = edgeCut(graph, part);

    while ((nbbalance < 2) && partSizeChecker(partsize,bound,nbpart, graph->maxVW)){
      previous_maxpartsize = maxpartsize;

        forcedBalance(graph, toporder, bound, part, nbpart, opt, info, partsize);
        info->current_edgecut = objvalue = edgeCut(graph, part);
		nbbalance++;
        randTopSortOnParts(graph, part, toporder, nbpart);
    }

    lastobjvalue = objvalue;

    info->ec_afterforced_tab[level] = info->current_edgecut;
    info->timing_forced_tab[level] += u_wseconds();

    if (opt.debug > 0){
		printf("We ran %d Forced Balance iterations\n", nbbalance);
		for(i=0;i<nbpart;++i)
			printf("%d, ",partsize[i] );
		printf("PARTSIZES\n");
    }

    if(objvalue < 0)
        u_errexit("The objective value has not been set correctly\n");

    info->nbref_tab[level] = 0;

    for (i=1; i<= opt.ref_step; i++) {
        if (opt.debug > 10){
            print_graph(graph, part, opt.file_name_dot, level, i);
        }
        info->timing_refinement_tab_tab[level][i] -= u_wseconds();

        switch(opt.refinement) {
            case REF_kFM_PO :
                refinementPostOrder(graph, toporder, opt.lb, opt.ub, part, nbpart, &objvalue, opt);
                break;

            case REF_kFM_TB :
                buildPartmatrix(graph, part, opt.nbPart, partmatrix);
                refinementTowardBest(graph, partmatrix, opt.lb, opt.ub, part, nbpart, &objvalue, opt);
                break;

            case REF_kFM_POTB :
                randTopSortOnParts(graph, part, toporder, nbpart);
                refinementPostOrder(graph, toporder, opt.lb, opt.ub, part, nbpart, &objvalue, opt);
                info->current_edgecut = objvalue;
                buildPartmatrix(graph, part, opt.nbPart, partmatrix);
                refinementTowardBest(graph, partmatrix, opt.lb, opt.ub, part, nbpart, &objvalue, opt);
                break;

            case REF_kFM_RANDV:
                buildPartmatrix(graph, part, opt.nbPart, partmatrix);
                refinementRandomizedSelectTowardsBest(graph, partmatrix, opt.lb, opt.ub, part, nbpart, &objvalue, opt);
                break;

            default:
                u_errexit("Refinement %d is not recognized!\n", opt.refinement);
        }

        info->current_edgecut = objvalue;
        info->timing_refinement_tab_tab[level][i] += u_wseconds();
        info->ec_afterref_tab[level][i] = info->current_edgecut;
        info->nbref_tab[level] = i;

        if(opt.debug>4) {
            printf("\tVerify edge cut for coars level %d refinement step %d (stored ec %.2f)\n", level, i, 1.0*info->current_edgecut);
            ecType cutComputed = edgeCut(graph, part);
            if(cutComputed !=  info->current_edgecut )
                u_errexit("edge cut computer %.2f but stored %.2f\n", 1.0*cutComputed, 1.0*info->current_edgecut );
        }
        if (opt.debug > 0)
            printf("   ref %d\t%d\t%d\t%d\t%d\n", i, graph->nVrtx, graph->nEdge, (int) objvalue, 0);

        if (objvalue >= 0.99*lastobjvalue)
            break;

        lastobjvalue = objvalue;
    }

    switch(opt.refinement) {
        case REF_kFM_TB :
        case REF_kFM_POTB :
        case REF_KL :
        case REF_KL_bFM_MAXW :
        case REF_kFM_RANDV:
            for (j = 0; j < opt.nbPart; j++)
                free(partmatrix[j]);
            free(partmatrix);
            break;
    }

    free(toporder);
    free(bound);
    free(partsize);
}
