#include <stdio.h>
#include <stdlib.h>
#include <float.h>

#include "vcycle.h"
#include "option.h"
#include "utils.h"
#include "dgraphReader.h"
#include "debug.h"
#include "clustering.h"
#include "dgraphPartition.h"
#include "initialPartitioning.h"
#include "initialBisection.h"
#include "refinement.h"
#include "refinementBis.h"
#include "info.h"

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wformat"
#pragma GCC diagnostic ignored "-Wunused-parameter"



coarsen *VCycle(dgraph *igraph, idxType* flag, const MLGP_option opt, MLGP_info* info)
{
    /*Proceed a vcycle starting from graph*/
    info->timing_global -= u_wseconds();
    idxType i;

    coarsen *current, *head = initializeCoarsen(igraph);

    int level = 0;
    idxType old_nbnode = igraph->nVrtx;

    current = head;

    for (i = 1; i <= igraph->nVrtx; i++){
        current->flag[i] = flag[i];
    }

    info->timing_coars -= u_wseconds();
    while ( (current->graph->nVrtx > opt.co_stop_size) &&  (level < opt.co_stop_level)) {
        if (opt.print >= PD_HIGH) {
            dgraph *g=current->graph;
            printf("Coarsening:%2d: G(%6d, %6d)  maxDeg= <%3d, %3d>   maxW= [%4d, %d]\n",
                level, g->nVrtx, g->nEdge, g->maxindegree, g->maxoutdegree, g->maxVW, (int) g->maxEC);
        }

        level++;
        info->timing_coars_tab[level] -= u_wseconds();
        info->timing_matching_tab[level] -= u_wseconds();
        computeClustering(opt, level, current, current->leader);
        idxType nodetoprint;
        dgraph* G = current->graph;

        info->timing_matching_tab[level] += u_wseconds();

        info->timing_buildgraph_tab[level] -= u_wseconds();
        dgraph* nextgraph = initializeNextGraph(current, opt);

        if (opt.toggle){
            printf("Ploting coars_%d.pdf ..... ", level);
            char fna[PATH_MAX];
            sprintf(fna, "plot/coars_%d.dot", level);
            dgraph_to_dot_with_matching(current->graph, current->leader, fna);
            sprintf(fna, "dot -Tpdf plot/coars_%d.dot -o plot/coars_%d.pdf", level, level);
            system(fna);
            printf("Done\n");
        }

        if (opt.debug >= PD_MEDIUM) {
            char fna[PATH_MAX];
            sprintf(fna, "plot/coars_%d.dot", level+1);
            dgraph_to_dot_with_matching(nextgraph, NULL, fna);

            printf("Check acyclicity depth %d level %d size %d edge %d\n", info->depth, level, nextgraph->nVrtx, nextgraph->nEdge);
            checkGraphAcyclicity(nextgraph);
            printf("Graph is acyclic\n");
        }

        info->nbnodes_coars_tab[level] = nextgraph->nVrtx;
        info->nbedges_coars_tab[level] = nextgraph->nEdge;
        coarsen* next = initializeCoarsen(nextgraph);
        for (i = 1; i <= G->nVrtx; i++)
            next->flag[current->new_index[i]] = current->flag[i];
        info->timing_buildgraph_tab[level] += u_wseconds();


        current->next_coarsen = next;
        next->previous_coarsen = current;

        if (opt.debug >= PD_HIGH) {
            char fn[PATH_MAX];
            sprintf(fn, "%s.CL.%02d.dot", opt.file_name, level);
            writeDGraphDot(next->graph, fn, NULL);
        }

        if (opt.debug > 100) {
           char fnn[100];

           sprintf(fnn,"matching_%d.txt", level);
           FILE *qq=fopen(fnn,"w");
           for (int i=1;i<=current->graph->nVrtx;++i) {
                if (i!=current->leader[i])
                    fprintf(qq, "%d %d\n", i, current->leader[i]);
           }
           fclose(qq);
        }

        current = next;
        info->timing_coars_tab[level] += u_wseconds();
        if ((double) current->graph->nVrtx / (double) old_nbnode > opt.co_stop_ratio)
            break;
        old_nbnode = current->graph->nVrtx;
    }
    if (opt.print >= PD_MEDIUM) {
        dgraph *g=current->graph;
        printf("Outside Coarsening Loop\nCoarsening:%2d: G(%6d, %6d)  maxDeg= <%3d, %3d>   maxW= [%4d, %d]\n",
            level, g->nVrtx, g->nEdge, g->maxindegree, g->maxoutdegree, g->maxVW, (int) g->maxEC);
    }
    info->coars_depth = level;
    info->timing_coars += u_wseconds();

    /*Initial partitioning*/
    info->timing_inipart -= u_wseconds();

    ecType coarsew = 0;
    for (int i=0; i<current->graph->nEdge; i++) {
        coarsew += current->graph->ecIn[i];
    }
    info->finalew = coarsew;

    initialPartitioning(opt, current, info);

    info->timing_inipart += u_wseconds();

    if (opt.debug > 0)
        printPartWeights(current->graph, current->part);

    if (opt.toggle){
        printf("Ploting part_%d.pdf ..... ", level+1);
        char fna[PATH_MAX];
        sprintf(fna, "plot/part_%d.dot", level+1);
        dgraph_to_dot(current->graph, current->part, fna);
        sprintf(fna, "dot -Tpdf plot/part_%d.dot -o plot/part_%d.pdf", level+1, level+1);
        system(fna);
        printf("Done\n");
    }

    info->timing_uncoars -= u_wseconds();

    /* Uncoarsening */
    for ( ; level>=0 ;  level--) {
        info->timing_uncoars_tab[level] -= u_wseconds();


        info->timing_refinement_tab[level] -= u_wseconds();
        KwayRefinementStep(opt, level, current->graph, current->part, current->nbpart, info);

        info->timing_refinement_tab[level] += u_wseconds();

        /* UVC TODO: refinment should update edgecut/nbcombb, partsizes and we should print them without recompute*/
        if (opt.print >= PD_HIGH) {
            ecType edgecut = edgeCut(current->graph, current->part);
            printf("Refinment:%2d: G(%6d, %6d) Cut=%d\n", level, current->graph->nVrtx, current->graph->nEdge, (int) edgecut);
        }

        if (current->previous_coarsen) {
            info->timing_project_tab[level] -= u_wseconds();
            coarsen *prev = current;
            current = projectBack(current);
            freeCoarsen(prev);
            info->timing_project_tab[level] += u_wseconds();
        }

        if (opt.toggle){
            printf("Ploting part_%d.pdf ..... ", level);
            char fna[PATH_MAX];
            sprintf(fna, "plot/part_%d.dot", level);
            dgraph_to_dot(current->graph, current->part, fna);
            sprintf(fna, "dot -Tpdf plot/part_%d.dot -o plot/part_%d.pdf", level, level);
            system(fna);
            printf("Done\n");
        }

        if (opt.debug > 0)
            printPartWeights(current->graph, current->part);


        info->timing_uncoars_tab[level] += u_wseconds();
    }
    if (opt.print > PD_NONE) {
        info->finalec = edgeCut(head->graph, head->part);
    }
    if (opt.print >= PD_MEDIUM) {
        printf("Refinment:Final: Cut=%d\n", (int) info->finalec);
    }

    info->timing_uncoars += u_wseconds();
    info->timing_global += u_wseconds();

    return head;

}
